@@ -1,21 +1,24 @@
 #!/usr/bin/env python3
+import sys, os, time, socket, re
 from netdevice import linux
 import simplejson as json
-import pexpect
-import xmltodict
 import ipaddress
+from binascii import hexlify, unhexlify
 import random
-from lxml import etree
 try:
     # Python 2.
     from StringIO import StringIO
     # Python 3.
 except ImportError:
     from io import StringIO
-import sys, os, time
-import collections
-import tempfile
-import copy
+
+try:
+    # Python 2.
+    import urlparse
+except ImportError:
+    # Python 3.
+    import urllib.parse as urlparse
+
 
 class OvsHost(linux.LinuxDevice):
     '''
@@ -30,6 +33,7 @@ class OvsHost(linux.LinuxDevice):
         '''
 
         linux.LinuxDevice.__init__(self, server, **kwargs)
+        self.devices = []
         self.macs = []
 
         # start the docker if it's not running
@@ -107,8 +111,8 @@ class OvsHost(linux.LinuxDevice):
         self.cmd(command) #execut the command.
 
         # delete the current offlow when it's created
-        #if self["fail-mode"] == "secure":
-        #    self.cmd("ovs-ofctl del-flows %s" %(name))
+        if self["fail-mode"] == "secure":
+            self.ofctl_add_flow(name, "priority=0,actions=drop")
 
         if ip:
             # Configure the ip address for the address for route
@@ -120,35 +124,42 @@ class OvsHost(linux.LinuxDevice):
         ports = ports if (isinstance(ports, list)) else [ports]
         ports = ports if (not args) else ports.extend(args)
         for p in ports:
-            self.add_port(name, **p)
+            self.add_port(name, p)
 
-    def add_port (self, bridge_name, name, **kwargs):
+    def add_port (self, bridge_name, kwargs):
         '''
         Add self-port and do some configuration if necessary.
         '''
+        name = kwargs.get("name", "tmpname")
         command = 'ovs-vsctl add-port %s %s' %(bridge_name, name)
-        ip = kwargs.pop("ip", []) #get the ip configuration of the bridge
         if kwargs:
             # If there is parameters, for example type=vxlan, etc.
             command += ' -- set interface %s' %(name)
             for k,v in kwargs.items():
-                command += ' %s=%s' %(k,v)
+                if (k not in ["name", "ip", "mac"]):
+                    command += ' %s=%s' %(k,v)
         self.cmd(command) #execut the command.
 
         #If it's a normal interface, flush it to release its original ip address.
         if (kwargs.get("type", "normal") == "normal"):
             self.cmd("ip addr flush dev %s" %(name))
 
-        if ip:
+        macaddr = self.ovs_get_interface(name, "mac_in_use").strip('"')
+        kwargs["mac"] = macaddr
+        if kwargs.get("ip", None) :
             # Configure the ip address for the address for route
             self.cmd('ip link set %s up' %(name))
-            self.cmd('ip address add %s dev %s' %(ip, name))
-            #self.cmd('ovs-appctl ovs/route/add %s %s' %(ip, name))
+            self.cmd('ip address add %s dev %s' %(kwargs["ip"], name))
+            #self.cmd('ovs-appctl ovs/route/add %s %s' %(kwargs["ip"], name))
             if self["fail-mode"] == "secure":
                 # Confiure the default offlow for self port with IP.
+                ipaddr = ipaddress.ip_interface(kwargs["ip"]).ip
                 self.ofctl_add_flow(bridge_name,
-                    "priority=65535,arp,arp_tpa=%s actions=normal" %(ip.split("/")[0]),
-                    "priority=1,ip,nw_dst=%s actions=normal" %(ip.split("/")[0]))
+                    #"priority=65535,arp,arp_tpa=%s,actions=normal" %(ip.split("/")[0]),
+                    #"priority=1,ip,nw_dst=%s actions=normal" %(ip.split("/")[0]),
+                    'priority=1,arp,arp_tpa=%s,arp_op=1,actions=move:"NXM_OF_ETH_SRC[]->NXM_OF_ETH_DST[]",mod_dl_src:"%s",load:"0x02->NXM_OF_ARP_OP[]",move:"NXM_NX_ARP_SHA[]->NXM_NX_ARP_THA[]",load:"0x%s->NXM_NX_ARP_SHA[]",move:"NXM_OF_ARP_SPA[]->NXM_OF_ARP_TPA[]",load:"0x%s->NXM_OF_ARP_SPA[]",in_port'%(ipaddr, macaddr, re.sub(r":", "", macaddr), "".join(f"{i:02x}" for i in ipaddr.packed)),
+                    'priority=1,icmp,nw_dst=%s,icmp_type=8,icmp_code=0,actions=push:"NXM_OF_ETH_SRC[]",push:"NXM_OF_ETH_DST[]",pop:"NXM_OF_ETH_SRC[]",pop:"NXM_OF_ETH_DST[]",push:"NXM_OF_IP_SRC[]",push:"NXM_OF_IP_DST[]",pop:"NXM_OF_IP_SRC[]",pop:"NXM_OF_IP_DST[]",load:"0xff->NXM_NX_IP_TTL[]",load:"0->NXM_OF_ICMP_TYPE[]",in_port' %(ipaddr),
+                    )
 
     def add_device (self, bridge_name, *args, **kwargs):
         '''
@@ -167,21 +178,20 @@ class OvsHost(linux.LinuxDevice):
                     %(arg["name"], arg["interface"]), log_level = 4)
                 # be compatible with the old version.
                 arg["mac"] = macaddr.split()[1]
-                self.macs.append(arg["mac"])
                 arg["ofport"] = self.ovs_get_container_ofport(arg["name"])
+                self.macs.append(arg["mac"])
+                self.devices.append(arg)
 
                 # Configure offlow for the same subnet
                 if self["fail-mode"] == "secure":
                     ipaddr = ipaddress.ip_interface(arg["ip"])
                     self.ofctl_add_flow(bridge_name,
-                        "priority=65535,arp,arp_tpa=%s actions=output:%s"
+                        "priority=65535,arp,arp_tpa=%s,actions=output:%s"
                         %(ipaddr.ip, arg["ofport"]),
-                        "priority=60000,ip,nw_src=%s,nw_dst=%s actions=output:%s" 
+                        "priority=60000,ip,nw_src=%s,nw_dst=%s,actions=output:%s" 
                         %(ipaddr.network, ipaddr.ip, arg["ofport"]))
-
             elif dtype == "vm":
-                self.__add_vm(bridge_name, **arg)
-
+                self.__add_vm(bridge_name, arg)
             else:
                 self.log("device type %s is not supported now..." %(dtype))
 
@@ -218,7 +228,7 @@ class OvsHost(linux.LinuxDevice):
                     %(ovs_get_container_port(bridge_name, name = name),
                         kwargs["vlan"]))
 
-    def __add_vm (self, bridge_name, **kwargs):
+    def __add_vm (self, bridge_name, vm):
         '''
         # The following parameter are mandatory:
         vm11 = {"name": "vm11", "type": "vm", "password": "sangfor",
@@ -229,27 +239,27 @@ class OvsHost(linux.LinuxDevice):
                 "port": [vm11_port0, vm11_port1]}
         '''
 
-        kwargs["-name"] = kwargs.get("-name", kwargs.get("name", "tmp"))
-        kwargs["-cpu"] = kwargs.get("-cpu", "host")
-        kwargs["-m"] = kwargs.get("-m", "2048M")
-        kwargs["-smp"] = kwargs.get("-smp", "sockets=1,cores=1")
-        kwargs["-boot"] = kwargs.get("-boot", "c")
-        kwargs["-pidfile"] = kwargs.get("-pidfile",
-                "/tmp/%s.pid" %(kwargs["-name"]))
-        kwargs["-enable-kvm"] = kwargs.get("-enable-kvm", "")
-        kwargs["-object"] = kwargs.get("-object",
+        vm["-name"] = vm.get("-name", vm.get("name", "tmp"))
+        vm["-cpu"] = vm.get("-cpu", "host")
+        vm["-m"] = vm.get("-m", "2048M")
+        vm["-smp"] = vm.get("-smp", "sockets=1,cores=1")
+        vm["-boot"] = vm.get("-boot", "c")
+        vm["-pidfile"] = vm.get("-pidfile",
+                "/tmp/%s.pid" %(vm["-name"]))
+        vm["-enable-kvm"] = vm.get("-enable-kvm", "")
+        vm["-object"] = vm.get("-object",
             "memory-backend-file,id=mem,size=%s,mem-path=/dev/hugepages,share=on"
-            %(kwargs["-m"]))
-        kwargs["-numa"] = kwargs.get("-numa", "node,memdev=mem")
-        kwargs["-mem-prealloc"] = kwargs.get("-mem-prealloc", "")
+            %(vm["-m"]))
+        vm["-numa"] = vm.get("-numa", "node,memdev=mem")
+        vm["-mem-prealloc"] = vm.get("-mem-prealloc", "")
 
-        cmd = kwargs.get("mask", None) and ("taskset %s " %(kwargs["mask"])) or ""
+        cmd = vm.get("mask", None) and ("taskset %s " %(vm["mask"])) or ""
         cmd += "qemu-system-x86_64"
-        for k,v in kwargs.items():
+        for k,v in vm.items():
             if k[0] == "-":
                 cmd += ' %s %s' %(k,v)
 
-        for p in kwargs.get("port", []):
+        for p in vm.get("port", []):
             for k,v in p.items():
                 if k[0] == "-":
                     #bypass the parameters: name, ip
@@ -286,22 +296,22 @@ class OvsHost(linux.LinuxDevice):
                 #print("netdev: %s" %(netdev))
                 #print("device: %s" %(device))
                 p["mac"] = device["mac"] # Record the mac for use later
-                port_name = "to_%s_%s" %(kwargs.get("name", "tmp"), netdev["id"])
+                port_name = "to_%s_%s" %(vm.get("name", "tmp"), netdev["id"])
                 ovs_port = {"name": port_name,
                         "type": "dpdkvhostuserclient",
                         "options:vhost-server-path": chardev["path"]}
-                self.add_port(bridge_name, **ovs_port)
+                self.add_port(bridge_name, ovs_port)
                 self.macs.append(p["mac"])
+                self.devices.append(p)
                 # Configure offlow for the same subnet
                 if self["fail-mode"] == "secure":
                     p["ofport"] = self.ovs_get_interface(port_name, "ofport")
                     ipaddr = ipaddress.ip_interface(p["ip"])
                     self.ofctl_add_flow(bridge_name,
-                        "priority=65535,arp,arp_tpa=%s actions=output:%s"
+                        "priority=65535,arp,arp_tpa=%s,actions=output:%s"
                         %(ipaddr.ip, p["ofport"]),
-                        "priority=60000,ip,nw_src=%s,nw_dst=%s actions=output:%s" 
+                        "priority=60000,ip,nw_src=%s,nw_dst=%s,actions=output:%s" 
                         %(ipaddr.network, ipaddr.ip, p["ofport"]))
-
         tmp = self.cmd("%s -daemonize" %(cmd))
         #val = os.system(cmd + "&")
 
@@ -310,18 +320,33 @@ class OvsHost(linux.LinuxDevice):
         Add some flows to ofproto
         '''
         vxlan_port = None # Get the vxlan at first.
+        #print("devices: %s" %(self.devices))
+        #print("mac: %s" %(self.macs))
+        if (gw["mac"] not in self.macs):
+            # gw is not on this node, Add it as if it's on this node.
+            ipaddr = ipaddress.ip_interface(gw["ip"]).ip
+            self.ofctl_add_flow(bridge_name,
+                    'priority=1,arp,arp_tpa=%s,arp_op=1,actions=move:"NXM_OF_ETH_SRC[]->NXM_OF_ETH_DST[]",mod_dl_src:"%s",load:"0x02->NXM_OF_ARP_OP[]",move:"NXM_NX_ARP_SHA[]->NXM_NX_ARP_THA[]",load:"0x%s->NXM_NX_ARP_SHA[]",move:"NXM_OF_ARP_SPA[]->NXM_OF_ARP_TPA[]",load:"0x%s->NXM_OF_ARP_SPA[]",in_port'%(ipaddr,
+                        gw["mac"],
+                        re.sub(r":", "", gw["mac"]),
+                        "".join(f"{i:02x}" for i in ipaddr.packed)),
+                    'priority=1,icmp,nw_dst=%s,icmp_type=8,icmp_code=0,actions=push:"NXM_OF_ETH_SRC[]",push:"NXM_OF_ETH_DST[]",pop:"NXM_OF_ETH_SRC[]",pop:"NXM_OF_ETH_DST[]",push:"NXM_OF_IP_SRC[]",push:"NXM_OF_IP_DST[]",pop:"NXM_OF_IP_SRC[]",pop:"NXM_OF_IP_DST[]",load:"0xff->NXM_NX_IP_TTL[]",load:"0->NXM_OF_ICMP_TYPE[]",in_port' %(ipaddr))
 
         for d in args:
             if d.get("type", None) == "vxlan":
                 vxlan_port = self.ovs_get_interface(d["name"], "ofport")
                 break
+
         for d in args:
+            #self.log("d: %s" %(d))
             if d.get("type", None) != "vxlan":
                 if ((d["mac"] not in self.macs) and (not vxlan_port)):
                     # d is not on host, but there is no vxlan given.
                     self.log("Error: there is no vxlan to remote device: %s!"
                             %(d["name"]), bg_color = "red")
                     return
+
+                #print("d: %s" %(d))
                 self.ofctl_add_flow(bridge_name,
                     "priority=10,ip,nw_dst=%s,action=mod_dl_src:%s,"
                     "mod_dl_dst:%s,dec_ttl,output:%s"
@@ -340,11 +365,53 @@ class OvsHost(linux.LinuxDevice):
             ip = ipaddress.ip_interface(arg["ip"])
             vxlan_port = self.ovs_get_interface(vtep["name"], "ofport")
             self.ofctl_add_flow(bridge_name,
-                    "priority=65535,arp,arp_tpa=%s actions=output:%s"
+                    "priority=65535,arp,arp_tpa=%s,actions=output:%s"
                     %(ip.ip, vxlan_port),
-                    "priority=60000,ip,nw_src=%s,nw_dst=%s actions=output:%s" 
+                    "priority=60000,ip,nw_src=%s,nw_dst=%s,actions=output:%s" 
                     %(ip.network, ip.ip, vxlan_port))
 
+    def set_out_interface (self, bridge_name, out_if, *args):
+        '''
+        Add some flows remote devices(*args): 从本主机上到达所有*args的包，全
+        部经由out_if
+        '''
+        if self["fail-mode"] != "secure":
+            #Don't need configure out_if explicitly in standalone mode.
+            return
+        for arg in args:
+            ip = ipaddress.ip_interface(arg["ip"])
+            vxlan_port = self.ovs_get_interface(out_if["name"], "ofport")
+            self.ofctl_add_flow(bridge_name,
+                    "priority=65535,arp,arp_tpa=%s,actions=output:%s"
+                    %(ip.ip, vxlan_port),
+                    "priority=60000,ip,nw_src=%s,nw_dst=%s,actions=output:%s" 
+                    %(ip.network, ip.ip, vxlan_port))
+
+    def set_next_hop (self, bridge_name, gw, *args):
+        '''
+        Add some flows to ofproto
+        '''
+        vxlan_port = None # Get the vxlan at first.
+
+        for d in args:
+            if d.get("type", None) == "vxlan":
+                vxlan_port = self.ovs_get_interface(d["name"], "ofport")
+                break
+        for d in args:
+            if d.get("type", None) != "vxlan":
+                if ((d["mac"] not in self.macs) and (not vxlan_port)):
+                    # d is not on host, but there is no vxlan given.
+                    self.log("Error: there is no vxlan to remote device: %s!"
+                            %(d["name"]), bg_color = "red")
+                    return
+
+                self.ofctl_add_flow(bridge_name,
+                    "priority=10,ip,nw_dst=%s,action=mod_dl_src:%s,"
+                    "mod_dl_dst:%s,dec_ttl,output:%s"
+                    %(d["ip"].split("/")[0], gw["mac"], d["mac"],
+                        (d["mac"] in self.macs) and d["ofport"] or vxlan_port))
+
+
     def del_br (self, *args, **kwargs):
         '''
         Delete the bridge and all the connected devices(containers).
@@ -431,7 +498,7 @@ class OvsHost(linux.LinuxDevice):
                 # remove the line starting with '#'
                 l = l.strip()
                 if l[0] !=  "#":
-                    self.cmd('ovs-ofctl add-flow %s "%s"' %(bridge_name, l))
+                    self.cmd('ovs-ofctl add-flow %s %s' %(bridge_name, l))
         return None
 
     def ping_test (self, src, dst):
